/*
 * ================================================================================
 * VoiceTest
 * Copyright (C) 2010 Spencer "voogru" MacDonald.  All rights reserved.
 * ================================================================================
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, version 3.0, as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "extension.h"
#include "filesystem.h"

VoiceTest g_VoiceTest;

SMEXT_LINK(&g_VoiceTest);

ICvar *g_pLocalCVar = NULL;
IFileSystem *g_pFileSystem = NULL;
IVoice *g_pVoice = NULL;

#define MAX_WAVEFILEDATA_LEN 1024 * 1024
typedef struct TwaveFileData {
	char *pDecompressedFileData;
	int nDecompressedDataBytes;
	const char *pDecompressedDataFilename;
} waveFileData;

waveFileData g_waveFileData[64];

// These are used to record all voice chat on the server.
char *g_pDecompressedFileData;
int g_nDecompressedDataBytes;
const char *g_pDecompressedDataFilename;

WAVEFORMATEX g_VoiceSampleFormat =
{
	WAVE_FORMAT_PCM,	// wFormatTag
	1,					// nChannels
	11025,				// nSamplesPerSec
	22050,				// nAvgBytesPerSec
	2,					// nBlockAlign
	16,					// wBitsPerSample
	0					// cbSize
};

CON_COMMAND(voicehook_start, "") {
	if (args.ArgC() < 2) {
		g_SMAPI->ConPrint("Usage: voicehook_start <client index> <filename>\n");
		return;
	}

	int clientIndex = atoi(args[1])-1;
	if (!g_waveFileData[clientIndex].pDecompressedFileData) {
		g_waveFileData[clientIndex].pDecompressedFileData = new char[MAX_WAVEFILEDATA_LEN];
		g_waveFileData[clientIndex].nDecompressedDataBytes = 0;
		g_waveFileData[clientIndex].pDecompressedDataFilename = args[2];
	} else {
		g_SMAPI->ConPrintf("Error: Already recording voice stream for client %i.\n", clientIndex+1);
		return;
	}
}

CON_COMMAND(voicehook_stop, "") {
	if (args.ArgC() < 1) {
		g_SMAPI->ConPrint("Usage: voicehook_stop <client index>\n");
		return;
	}

	int clientIndex = atoi(args[1])-1;
	if(g_waveFileData[clientIndex].pDecompressedFileData) {
		WriteWaveFile(g_waveFileData[clientIndex].pDecompressedDataFilename, g_waveFileData[clientIndex].pDecompressedFileData, g_waveFileData[clientIndex].nDecompressedDataBytes, g_VoiceSampleFormat.wBitsPerSample, g_VoiceSampleFormat.nChannels, g_VoiceSampleFormat.nSamplesPerSec);
		delete[] g_waveFileData[clientIndex].pDecompressedFileData;
		g_waveFileData[clientIndex].pDecompressedFileData = NULL;
	} else {
		g_SMAPI->ConPrintf("Error: Not currently recording voice stream for client %i.\n", clientIndex+1);
		return;
	}
}

void VoiceTest::OnVoiceBroadcast(int client, char *pVoiceData, int *nSamples)
{
	if(g_waveFileData[client-1].pDecompressedFileData)
	{
		g_SMAPI->ConPrintf("Recording %d samples of voice data for client %d.\n", *nSamples, client);

		int nToWrite = min(*nSamples * 2, MAX_WAVEFILEDATA_LEN - g_waveFileData[client-1].nDecompressedDataBytes);
		memcpy(&g_waveFileData[client-1].pDecompressedFileData[g_waveFileData[client-1].nDecompressedDataBytes], pVoiceData, nToWrite);
		g_waveFileData[client-1].nDecompressedDataBytes += nToWrite;
	}
}

bool VoiceTest::RegisterConCommandBase(ConCommandBase *pCommandBase)
{
	return META_REGCVAR(pCommandBase);
}

bool VoiceTest::SDK_OnLoad(char *error, size_t maxlen, bool late)
{
	return true;
}

bool VoiceTest::SDK_OnMetamodLoad(ISmmAPI *ismm, char *error, size_t maxlen, bool late)
{
	GET_V_IFACE_CURRENT(GetFileSystemFactory, g_pFileSystem, IFileSystem, FILESYSTEM_INTERFACE_VERSION);
	if (!g_pFileSystem)
	{
		snprintf(error, maxlen, "Could not find interface %s", FILESYSTEM_INTERFACE_VERSION);
		return false;
	}

	GET_V_IFACE_CURRENT(GetEngineFactory, g_pLocalCVar, ICvar, CVAR_INTERFACE_VERSION);
	if (!g_pLocalCVar)
	{
		snprintf(error, maxlen, "Could not find interface %s", CVAR_INTERFACE_VERSION);
		return false;
	}

	int ifaceerror;
	g_pVoice = (IVoice *)g_SMAPI->MetaFactory(VOICE_INTERFACE, &ifaceerror, NULL);
	if (!g_pVoice)
	{
		snprintf(error, maxlen, "Could not find interface %s (error %d)", VOICE_INTERFACE, ifaceerror);
		return false;
	}

	g_pVoice->AddVoiceListener(this);

	g_pCVar = g_pLocalCVar;
	ConVar_Register(0, this);

	return true;
}

bool VoiceTest::SDK_OnMetamodUnload(char *error, size_t maxlen)
{
	g_pVoice->RemoveVoiceListener(this);

	return true;
}

// Wave file code /////////////////////////////////////////////////////////////////////////
static unsigned long ReadDWord(FileHandle_t fp) 
{
	unsigned long ret;  
	g_pFileSystem->Read(&ret, 4, fp); 
	return ret;
}

static unsigned short ReadWord(FileHandle_t fp) 
{
	unsigned short ret; 
	g_pFileSystem->Read(&ret, 2, fp);
	return ret;
}

static void WriteDWord(FileHandle_t fp, unsigned long val) 
{
	g_pFileSystem->Write(&val, 4, fp);
}

static void WriteWord(FileHandle_t fp, unsigned short val) 
{
	g_pFileSystem->Write(&val, 2, fp);
}

bool WriteWaveFile(
	const char *pFilename, 
	const char *pData, 
	int nBytes, 
	int wBitsPerSample, 
	int nChannels, 
	int nSamplesPerSec)
{
	FileHandle_t fp = g_pFileSystem->Open(pFilename, "wb");
	if(!fp)
		return false;

	// Write the RIFF chunk.
	g_pFileSystem->Write("RIFF", 4, fp);
	WriteDWord(fp, 0);
	g_pFileSystem->Write("WAVE", 4, fp);


	// Write the FORMAT chunk.
	g_pFileSystem->Write("fmt ", 4, fp);

	WriteDWord(fp, 0x10);
	WriteWord(fp, WAVE_FORMAT_PCM);
	WriteWord(fp, (unsigned short)nChannels);	
	WriteDWord(fp, (unsigned long)nSamplesPerSec);
	WriteDWord(fp, (unsigned long)((wBitsPerSample / 8) * nChannels * nSamplesPerSec));
	WriteWord(fp, (unsigned short)((wBitsPerSample / 8) * nChannels));
	WriteWord(fp, (unsigned long)wBitsPerSample);

	// Write the DATA chunk.
	g_pFileSystem->Write("data", 4, fp);
	WriteDWord(fp, (unsigned long)nBytes);
	g_pFileSystem->Write(pData, nBytes, fp);


	// Go back and write the length of the riff file.
	unsigned long dwVal = g_pFileSystem->Tell(fp) - 8;
	g_pFileSystem->Seek(fp, 4, FILESYSTEM_SEEK_HEAD);
	WriteDWord(fp, dwVal);

	g_pFileSystem->Close(fp);
	return true;
}

bool ReadWaveFile(
	const char *pFilename,
	char *&pData,
	int &nDataBytes,
	int &wBitsPerSample,
	int &nChannels,
	int &nSamplesPerSec)
{
	FileHandle_t fp = g_pFileSystem->Open(pFilename, "rb");
	if(!fp)
		return false;

	g_pFileSystem->Seek(fp, 22, FILESYSTEM_SEEK_HEAD);

	nChannels = ReadWord(fp);
	nSamplesPerSec = ReadDWord(fp);

	g_pFileSystem->Seek(fp, 34, FILESYSTEM_SEEK_HEAD);
	wBitsPerSample = ReadWord(fp);

	g_pFileSystem->Seek(fp, 40, FILESYSTEM_SEEK_HEAD);
	nDataBytes = ReadDWord(fp);
	ReadDWord(fp);
	pData = new char[nDataBytes];
	if(!pData)
	{
		g_pFileSystem->Close(fp);
		return false;
	}
	g_pFileSystem->Read(pData, nDataBytes, fp);
	return true;
}
